Add LoadPath "CoqEAL".
Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq choice fintype.
Require Import finfun tuple ssralg matrix mxalgebra bigop.
Require Import definitions_auxlemmas inverse_nilpotent.
Require Import general_decomposition hexagonal_lemma  ZArith Zinfra.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Section BPL.

Open Local Scope ring_scope.


(*To apply the BPL, we need a reduction rho: D_rho => d_rho and a perturbation
  for every dimension whose size is the same that the differential of D_rho in the
  same dimension. This perturbation has to verify:
   1) the nilpotency differential: (D_rho (i+1) + P (i+1)) * (D_rho i + P i) = 0
   2) the nilpotency of the product of PH: exists n, (P i * H_rho i)^n = 0 *)
Section perturbation.

Variable K: fieldType.
Variable rho : FGReduction K.

Notation "'D_rho'" := (diff (C rho)).
Notation "'d_rho'" := (diff (D rho)).
Notation "'F_rho'" := (M (F rho)).
Notation "'G_rho'" := (M (G rho)).
Notation "'H_rho'" := (Ho (H rho)).
Notation "'ci'" := (m (C rho)).

Variable P :  forall i:Z, 'M[K]_(ci (i+1), ci i).
Variable (n : nat).

Hypothesis nilpotency_dp : forall i:Z, 
   (D_rho (i+1) + P (i+1)) *m (D_rho i + P i) = 0.
Hypothesis nilpotency_hp :  forall i:Z,
    (pot_matrix (P i *m H_rho i) n = 0).



(*We use the general decomposition given by the the isomorphisms Fi_isom and Gi_isom    to obtain the new perturbation and the new differential.
 *)
Section new_perturbation.

Definition Pi_new (i:Z):= (Fi_isom rho (i+1)) *m (P (i+1)) *m (Gi_isom rho i). 

Definition Di_new_pert(i:Z):=(Fi_isom rho (i+1)) *m (D_rho (i+1) + P(i+1)) *m (Gi_isom rho i).


(*The hypothesis 1) transferred to the decomposed chain complex*)
Lemma nilpotency_dp_new : forall i:Z, Di_new_pert (i+1) *m Di_new_pert i = 0.
Proof.
move=> i.
by rewrite /Di_new_pert -!mulmxA [(Gi_isom rho (i+1)) *m _]mulmxA (GiFi rho (i+1)) mul1mx [(D_rho (i+1+1) + _) *m _]mulmxA nilpotency_dp mul0mx mulmx0.
Qed.

(*Auxiliar lemma to prove: nilpotency_hp_new*)
Lemma pot_P1_new_H0_new k : forall i:Z,
   pot_matrix (Pi_new i *m (H_rho_change_base rho i)) k =
   (Fi_isom rho (i+1)) *m (pot_matrix (P (i+1) *m (H_rho (i+1))) k) *m
   (Gi_isom rho (i+1)).
Proof.
move=> i.
elim : k => [/=| k H]; first by rewrite mulmx1 (FiGi rho (i+1)).
have ->:  pot_matrix (Pi_new i *m H_rho_change_base rho i) k.+1 =
   (Pi_new i *m H_rho_change_base rho i) *m 
     pot_matrix (Pi_new i *m H_rho_change_base rho i) k by done.
rewrite H /Pi_new /H_rho_change_base -mulmxA -mulmxA -mulmxA -mulmxA -mulmxA
   [(Gi_isom rho i) *m _]mulmxA (GiFi rho i) mul1mx ![(Gi_isom rho (i+1)) *m _]mulmxA (GiFi rho (i+1)) mul1mx.
have -> : pot_matrix (P (i + 1) *m H_rho (i + 1)) k.+1 =  (P (i + 1) *m H_rho (i + 1)) *m pot_matrix  (P (i + 1) *m H_rho (i + 1)) k by done.
by rewrite !mulmxA.
Qed.

(*The property 2) transferred to the decomposed chain complex  *)
Lemma nilpotency_hp_new : forall i:Z,
 pot_matrix ((Pi_new i) *m (H_rho_change_base rho i)) n = 0.
Proof.
move=> i.
move : (nilpotency_hp (i+1))=> IH1.
by rewrite (pot_P1_new_H0_new n i) IH1 mulmx0 mul0mx.
Qed.



(*In the next section, we want to prove that a block of the Di_new_pert is invertible   to be in the conditions of applying the hexagonal lemma*)

(*Di_new_pert_ab = D_rho_change_base_ab + Pi_new_ab =
                 = D_aux + Pi_new_ab
                 = D_aux + (D_aux * H_aux) * Pi_new_ab  
                     (by the general decomposition (D_aux * H_aux = Id))
                 = D_aux (Id + H_aux * Pi_new_ab)


Then Di_new_pert_ab has inverse if:
  a) D_aux has inverse  
  b) (Id + H_aux * Pi_new_ab) has inverse

  a) We know that the inverse of D_aux is H_aux 
          (since by teh general decomposition (D_aux * H_aux = Id)))
  b) The inverse of (Id + H_aux * Pi_new_ab) is obtained by a lemma proved in 
     the file inverse_nilpotent where it is given the inverse of this matrix if
     the condition exists n, (H_aux * Pi_new_ab)^n = 0

Then, we want to prove that exists n, (H_aux * Pi_new_ab)^n = 0
*)

Section condition_BPL.

(*By hypotheses 2), we have that (Pi_new*H_rho_change_base)^n = 0.
  Then (a block of this product)^n is 0. This corresponds with the lemma H0_aux_P11.
*)

Lemma pot_block_a_0_b_0 (p1 p2 k: nat) (a : 'M[K]_p1)
   (b : 'M[K]_(p2,p1)) : 
  ulsubmx (pot_matrix (block_mx a 0 b 0) k) = pot_matrix a k.
Proof.
elim : k =>  [/=| k H]; first by rewrite scalar_mx_block block_mxKul.
have -> : (pot_matrix (block_mx a 0 b 0) k.+1) =
 (block_mx a 0 b 0) *m (pot_matrix (block_mx a 0 b 0) k) by done.
by rewrite -(submxK (pot_matrix (block_mx a 0 b 0) k)) mulmx_block
        block_mxKul mul0mx H GRing.addr0.
Qed.


Lemma H0_aux_P11 : forall i:Z,
  pot_matrix ((lsubmx (ursubmx(Pi_new i))) *m (H_aux rho i)) n = 0.
Proof.
move=> i.
move : (nilpotency_hp_new i).
rewrite -{1}(submxK (Pi_new i))-{1}(hsubmxK (ursubmx (Pi_new i)))
   -{1}(vsubmxK (dlsubmx (Pi_new i))) -{1}(submxK (drsubmx (Pi_new i))).
rewrite (H_rho_change_baseP rho i) mulmx_block -!block_mx0  mul_row_col mul_block_col !mulmx0  !row_mx0 !mulmx0 !GRing.add0r !GRing.addr0 block_mx0 block_mx0  !mulmx0
 -(submxK (pot_matrix
  (block_mx (lsubmx (ursubmx (Pi_new i)) *m H_aux rho i) 0
        (col_mx (ulsubmx (drsubmx (Pi_new i)) *m H_aux rho i)
           (dlsubmx (drsubmx (Pi_new i)) *m H_aux rho i)) 0) n)).
move => H.
move : (eq_block_mx H) => [HH _  _ _].
by move: HH; rewrite pot_block_a_0_b_0.
Qed.


(*Thanks to the lemma proved in the file inverse nilpotent, we can define that
  the inverse of Di_new_pert_ab = D_aux (Id + H_aux * Pi_new_ab) is
    H_aux * \sum (-Pi_new_ab * H_aux)^j
*)

Definition inverse_d1p1_11 (i:Z) :=  
 (H_aux rho i) *m (\sum_(0<=j<n) (pot_matrix (- (lsubmx (ursubmx(Pi_new i)) *m (H_aux rho i)))) j).


(*Then we prove the two lemmas corresponding with the block of the differential is
invertible: 
  Di_new_pert_ab * inverse_d1p1_11 = Id.
  inverse_d1p1_11 * Di_new_pert_ab = Id.
*)

Lemma inverse_d1p1_11_is_inverse :forall i:Z,
 (lsubmx (ursubmx (Di_new_pert (i+1)))) *m (inverse_d1p1_11 (i+
1)) = 1%:M.  
Proof.
move => i.
rewrite /Di_new_pert.
have <-: (D_rho_change_base rho (i + 1) + Pi_new (i + 1)) = 
       (Fi_isom rho (i+1+1)) *m (D_rho (i+1+1) + (P (i+1+1))) *m (Gi_isom rho (i+1)); first by rewrite /D_rho_change_base /Pi_new -mulmx_addl -mulmx_addr.
rewrite -(submxK (D_rho_change_base rho (i+1))) -{1}(hsubmxK (ursubmx(D_rho_change_base rho (i+1)))) -{1}(vsubmxK (dlsubmx (D_rho_change_base rho (i+1)))) -{1}(submxK (drsubmx (D_rho_change_base rho (i+1)))) -{1}(submxK (Pi_new (i+1))) -{1}(hsubmxK (ursubmx (Pi_new (i+1)))) -{1}(vsubmxK (dlsubmx (Pi_new (i+1)))) -{1}(submxK (drsubmx (Pi_new (i+1)))) add_block_mx  block_mxKur add_row_mx row_mxKl.
move : (D_rho_change_baseP rho (i+1)) => ->.
rewrite block_mxKur row_mxKl -(mulmx1 (lsubmx (ursubmx (Pi_new (i + 1)))))
  -{1}(H_aux_D_aux rho i) -{1}(mul1mx (D_aux rho (i+1))) -{1}(D_aux_H_aux rho (i+1)) 
  mulmxA -mulmx_addl /inverse_d1p1_11 -mulmxA.
have -> :  (D_aux rho (i + 1) *m (H_aux rho (i + 1) *m \big[  +%R/0]_(0 <= j < n)
         pot_matrix (- (lsubmx (ursubmx (Pi_new (i + 1))) *m  H_aux rho  (i + 1))) j))  = ((D_aux rho (i + 1) *m H_aux rho (i + 1)) *m \big[ +%R/0]_(0 <= j < n)
       pot_matrix (- (lsubmx (ursubmx (Pi_new (i + 1))) *m  H_aux rho  (i + 1))) j); 
  first by rewrite mulmxA.
by rewrite (D_aux_H_aux rho (i+1)) mul1mx (inverse_I_plus_M_big' (H0_aux_P11 (i+1))).
Qed.


Lemma inverse_d1p1_11_is_inverse' :forall i:Z,
 (inverse_d1p1_11 (i+1))*m  (lsubmx (ursubmx (Di_new_pert (i+1)))) = 1%:M.  
Proof.
move => i.
rewrite /Di_new_pert.
have <-: (D_rho_change_base rho (i + 1) + Pi_new (i + 1)) = 
       (Fi_isom rho (i+1+1)) *m (D_rho (i+1+1) + (P (i+1+1))) *m (Gi_isom rho (i+1))
  by rewrite /D_rho_change_base /Pi_new -mulmx_addl -mulmx_addr.
rewrite  -(submxK (D_rho_change_base rho (i+1))) -{1}(hsubmxK (ursubmx(D_rho_change_base rho (i+1)))) -{1}(vsubmxK (dlsubmx (D_rho_change_base rho (i+1)))) -{1}(submxK (drsubmx (D_rho_change_base rho (i+1)))) -{1}(submxK (Pi_new (i+1))) -{1}(hsubmxK (ursubmx (Pi_new (i+1)))) -{1}(vsubmxK (dlsubmx (Pi_new (i+1)))) -{1}(submxK (drsubmx (Pi_new (i+1)))) add_block_mx block_mxKur add_row_mx row_mxKl.
move : (D_rho_change_baseP rho (i+1)) => ->.
by rewrite  block_mxKur row_mxKl -(mulmx1 (lsubmx (ursubmx (Pi_new (i + 1)))))-{1}(H_aux_D_aux rho i) -{1}(mul1mx (D_aux rho (i+1)))  mulmxA -mulmx_addl /inverse_d1p1_11 -mulmxA  [_ *m (_ *m (D_aux rho (i+1)))]mulmxA (inverse_I_plus_M_big (H0_aux_P11 (i+1))) mul1mx (H_aux_D_aux rho i).
Qed.


Lemma inverse_d1p1_11_inverse : forall i:Z,
   (lsubmx (ursubmx (Di_new_pert (i+1)))) *m (inverse_d1p1_11 (i+
1)) = 1%:M 
  /\(inverse_d1p1_11 (i+1))*m  (lsubmx (ursubmx (Di_new_pert (i+1)))) = 1%:M. 
Proof.
by split; rewrite ?inverse_d1p1_11_is_inverse' ?inverse_d1p1_11_is_inverse.
Qed.


End condition_BPL.

(*Finally, we can apply the hexagonal lemma (the reduction is moved one
       dimension).
  Then, we obtain a reduction of the decomposed chain complex (Di_new_pert)
  with the pertubation Pi_new to another one (which will the cc (D quasi_bpl))
  but moved two dimensions.
*)

Definition quasi_bpl :=
  (rhoHe (Di:=Di_new_pert) (nilpotency :=nilpotency_dp_new) inverse_d1p1_11_inverse).



(*But, we want a reduction between the initial chain complex D_rho and this reduced 
  (D quasi_bpl). To be able to use the previous reduction, we need to define the 
  differentials of the initial chain complex D_rho perturbed in two dimensions more.

  Then, Di_BPL is the sum of the differential D_rho (two dimensions more) and the 
  perturbation two dimensions more.

  Using the isomorphism between D_rho and the decomposed one we define the new 
   morphisms and the new homotopy operator.

  On the other hand, we have that the chain complex reduced obtained by the 
  quasi_bpl will be the reduced chain complex obtained applying BPL with the initial
*)

Section definitions_BPL.

Definition Di_BPL (i:Z) := (D_rho (i+1+1) + P (i+1+1)).

Definition Di'_BPL (i:Z) := (diff (D quasi_bpl) i). 


(* ci:=m (C rho).*)

Definition Fi_BPL (i:Z) := 
  (Gi_isom rho (i+1)*m(M (F quasi_bpl) (i))).

Definition Gi_BPL (i:Z) := 
 (M (G quasi_bpl) (i)) *m Fi_isom  rho (i+1).


Definition Hi_BPL(i:Z) := 
  Gi_isom rho (i+1) *m ((Ho (H quasi_bpl)) (i))*m Fi_isom rho (i+1+1).


(*It is proved the properties of the reduction from the chain complex composed by
  Di_BPL to the chain complex composed by Di'_BPL*)
Section reduction_properties.



(*Property: the differential Di_BPL builds a chain complex *)
Lemma ddi_BPL : forall i,  Di_BPL (i+1) *m Di_BPL i = 0.
Proof. 
move=>i.
by rewrite nilpotency_dp.
Qed.


(*properties of a reduction*)
Lemma fgi_BPL : forall i, Gi_BPL i *m Fi_BPL i = 1%:M.
Proof.
move => i.
 rewrite /Gi_BPL /Fi_BPL -mulmxA.
 rewrite [(Fi_isom rho (i+1)) *m _]mulmxA (FiGi rho (i+1)) mul1mx.
 by rewrite (ax1 quasi_bpl i). 
Qed.


Lemma fhi_BPL :forall i,  Hi_BPL i *m Fi_BPL (i+1) = 0.
Proof.
move=>i.
by rewrite /Hi_BPL /Fi_BPL -mulmxA [(Fi_isom rho (i+1+1)) *m _]mulmxA 
  (FiGi rho (i+1+1))  mul1mx -mulmxA (ax3 quasi_bpl i) mulmx0.
Qed.



Lemma hgi_BPL : forall i, Gi_BPL i *m Hi_BPL i = 0.
Proof.
move=> i.
by rewrite /Hi_BPL /Gi_BPL -mulmxA ![(Fi_isom rho (i+1)) *m _]mulmxA(FiGi rho (i+1))
  mul1mx mulmxA (ax4 quasi_bpl i) mul0mx.
Qed.


Lemma hhi_BPL :forall i, Hi_BPL i *m Hi_BPL (i+1) = 0.
Proof.
move=> i.
by rewrite /Hi_BPL -mulmxA ![(Fi_isom rho (i+1+1)) *m _]mulmxA (FiGi rho (i+1+1)) 
  mul1mx -mulmxA
       [_ *m (_ *m (Fi_isom rho (i+1+1+1)))]mulmxA
       (ax5 quasi_bpl i) mul0mx mulmx0.
Qed.


Lemma gf_Dh_hDi_BPL :forall i,  Fi_BPL (i+1) *m Gi_BPL (i+1) + Hi_BPL (i+1) *m Di_BPL (i+1)  + Di_BPL(i) *m Hi_BPL (i) = 1%:M.
Proof.
move=> i.
rewrite /Fi_BPL /Gi_BPL /Hi_BPL /Di_BPL.
have ->: (D_rho (i + 1 + 1+1) + P (i + 1 + 1+1)) = (Gi_isom rho (i+1+1+1)) *m 
   (diff (C quasi_bpl) (i+1)) *m (Fi_isom rho (i+1+1)).
  have ->:  diff (C quasi_bpl) (i + 1) = Di_new_pert (i+1+1) by done.
  by rewrite /= /Di_new_pert -!mulmxA (GiFi rho (i+1+1)) mulmx1 mulmxA 
  (GiFi rho (i+1+1+1)) mul1mx.
rewrite !mulmxA -mulmx_addl -!mulmxA -mulmx_addr [(Fi_isom rho (i+1+1+1)) *m (Gi_isom rho (i + 1 + 1 + 1) *m _)]mulmxA (FiGi rho (i+1+1+1)) mul1mx. 
have ->: (D_rho (i + 1 + 1) + P (i + 1 + 1)) = (Gi_isom rho (i+1+1)) *m 
   (diff (C quasi_bpl) (i)) *m (Fi_isom rho (i+1)). 
  have ->:  diff (C quasi_bpl) (i) = Di_new_pert (i+1) by done.
  by rewrite /= /Di_new_pert -!mulmxA (GiFi rho (i+1)) mulmx1 mulmxA 
  (GiFi rho (i+1+1)) mul1mx.
rewrite !mulmxA -mulmx_addl -!mulmxA -mulmx_addr [(Gi_isom rho (i+1)) *m (Fi_isom rho (i + 1) *m _)]mulmxA (GiFi rho (i+1)) mul1mx. 
have ->: Fi_isom rho (i + 1 + 1) *m ((D_rho (i + 1 + 1) + P (i + 1 + 1)) *m 
   (Gi_isom rho (i + 1) *m  Ho (H quasi_bpl) i)) =
    ((Fi_isom rho (i + 1 + 1) *m  (D_rho (i + 1 + 1) + P (i + 1 + 1)) *m 
    (Gi_isom rho (i + 1)))) *m  (Ho (H quasi_bpl) i) by rewrite !mulmxA.
have ->: (Fi_isom rho (i + 1 + 1 + 1) *m 
     ((D_rho (i + 1 + 1 + 1) + P (i + 1 + 1 + 1)) *m  Gi_isom rho (i + 1 + 1))) = 
    (Fi_isom rho (i + 1 + 1 + 1) *m 
  ((D_rho (i + 1 + 1 + 1) + P (i + 1 + 1 + 1))) *m   Gi_isom rho (i + 1 + 1)) by rewrite !mulmxA.
by rewrite (ax2 quasi_bpl (i)) mulmx1 (GiFi rho (i+1+1)).
Qed.



(*Morphisms properties Fi_BPL, Gi_BPL*)
Lemma df_fDi_BPL :forall i,  Di_BPL i *m Fi_BPL i =  Fi_BPL (i + 1) *m Di'_BPL i.
Proof. 
move=> i.
rewrite /Fi_BPL /Di'_BPL /Di_BPL -mulmxA -(M_well_defined (F quasi_bpl) i).
have -> :  (D_rho (i + 1 + 1) + P (i + 1 + 1)) = (Gi_isom rho (i+1+1)) *m 
   (diff (C quasi_bpl) (i)) *m (Fi_isom rho (i+1)). 
  have ->:  diff (C quasi_bpl) (i) = Di_new_pert (i+1) by done.
  by rewrite /= /Di_new_pert -!mulmxA (GiFi rho (i+1)) mulmx1 mulmxA
   (GiFi rho (i+1+1)) mul1mx.
by rewrite -!mulmxA [(Fi_isom rho (i+1)) *m _]mulmxA (FiGi rho (i+1)) mul1mx.
Qed.


Lemma Dg_gdi_BPL : forall i,  Di'_BPL i *m Gi_BPL i = Gi_BPL (i+1) *m (Di_BPL i).
Proof. 
move=> i.
rewrite /Gi_BPL /Di'_BPL /Di_BPL /Gi_BPL mulmxA (M_well_defined (G quasi_bpl) i).
have -> :  (D_rho (i + 1 + 1) + P (i + 1 + 1)) = (Gi_isom rho (i+1+1)) *m 
   (diff (C quasi_bpl) (i)) *m (Fi_isom rho (i+1)). 
  have ->:  diff (C quasi_bpl) (i) = Di_new_pert (i+1) by done.
  by rewrite /= /Di_new_pert -!mulmxA (GiFi rho (i+1)) mulmx1 mulmxA 
  (GiFi rho (i+1+1)) mul1mx.
by rewrite -!mulmxA [(Gi_isom rho (i+1+1)) *m _]mulmxA (GiFi rho (i+1+1)) mul1mx.
Qed.



(*We define the chain complex composed by Di_BPL, the morphisms f_BPL and
  g_BPL and the homotopy operator h_BPL and finally, the reduction rho_bpl
  from the chain complex composed by Di_BPL = D_rho + P *)
Definition Dii:= Build_FGChain_Complex ddi_BPL.


Definition f_BPL := Build_FGChain_Complex_Morphism (A:=Dii) (*(B:=newD'))*) df_fDi_BPL.

Definition g_BPL := Build_FGChain_Complex_Morphism (B:=Dii) Dg_gdi_BPL.

Definition h_BPL:= Build_FGHomotopy_operator (A:=Dii) Hi_BPL.

Definition rho_bpl := 
 (Build_FGReduction (F:=f_BPL)(G:=g_BPL)(H:=h_BPL)fgi_BPL gf_Dh_hDi_BPL fhi_BPL hgi_BPL hhi_BPL).

End reduction_properties.

End definitions_BPL.


End new_perturbation.

(*Let us remember that this reduction has been obtained from the reduction initial 
  rho but moved two dimensions more. Then, we have to move two minus dimensions to
  apply the BPL lemma over the reduction initial *)


Section definitions_def.

Lemma cast1:forall i, ci(i-1-1+1+1+1)=ci(i+1).
Proof.
move=> i.
by rewrite !subr_addr.
Qed.


Lemma cast2:forall i, ci(i-1-1+1+1)=ci(i).
Proof.
move=> i.
by rewrite !subr_addr.
Qed.


Lemma cast11:forall i, m (C rho) (i-1-1+1+1+1)= m (C rho)(i+1).
Proof.
move=> i.
by rewrite !subr_addr.
Qed.


Lemma cast21:forall i, m (C rho)(i-1-1+1+1)= m (C rho)(i).
Proof.
move=> i.
by rewrite !subr_addr.
Qed.




Lemma cast3:forall i, m(D rho) (i-1-1+1+1+1)=m (D rho) (i+1).
rewrite //=.
rewrite /m3'.
move=> i.
by rewrite !subr_addr.
Qed.


Lemma cast4:forall i, m (D rho) (i - 1 - 1 + 1 + 1) = m (D rho) i.
rewrite //=.
rewrite /m3'.
move=> i.
by rewrite !subr_addr.
Qed.


Lemma cast5:forall i, ci (i - 1 - 1 + 1 + 1+1+1) = ci (i + 1 + 1).
rewrite //=.
move=> i.
by rewrite !subr_addr.
Qed.

Lemma cast6:forall i, m (D rho) (i - 1 - 1 + 1 + 1+1+1) = m (D rho) (i + 1 + 1).
rewrite //=.
move=> i.
by rewrite !subr_addr.
Qed.

Lemma cast31:forall i, m(D quasi_bpl) (i-1-1+1+1+1)=m (D quasi_bpl) (i+1).
rewrite //=.
rewrite /m3'.
move=> i.
by rewrite !subr_addr.
Qed.


Lemma cast41:forall i, m (D quasi_bpl) (i - 1 - 1 + 1 + 1) = m (D quasi_bpl) i.
rewrite //=.
rewrite /m3'.
move=> i.
by rewrite !subr_addr.
Qed.


Lemma cast61:forall i, m (D quasi_bpl) (i - 1 - 1 + 1 + 1+1+1) = m (D quasi_bpl) (i + 1 + 1).
rewrite //=.
move=> i.
by rewrite !subr_addr.
Qed.


Lemma id_castmx (m m': nat) (eq_mm : (m = m') * (m = m'))
  (R : ringType) :
  castmx eq_mm (1%:M) =  (@scalar_mx R m' 1).
Proof. 
by case: eq_mm; case: m' / => b; rewrite castmx_id. 
Qed.


Lemma addmx_cast : forall (R : ringType) (m1 n1 m2 n2 : nat) (eq_mn1 : (m1 = m2) * (n1 = n2))
     (A : 'M[R]_(m1, n1)) (B :'M[R]_(m1,n1)), 
     ((castmx eq_mn1 A) + (castmx eq_mn1 B)) = (castmx eq_mn1 (A + B)).
Proof.
move => R m1 n1 m2 n2 [eq_mn1 eq_mn1'] A B.
move: (eq_mn1) (eq_mn1') A B. 
rewrite eq_mn1 eq_mn1' => eq_mn_0 eq_mn_1 A B.
by rewrite !castmx_id.
Qed.




(*We define the differentials, the morphisms and the homotopy operator in 
  two dimensions less*)

Definition Di_def (i:Z):= castmx(cast1 i,cast2 i) (Di_BPL(i-1-1)).

Lemma equal_CC_def : forall i, ((Di_def i) = ((D_rho i) + (P i))).
Proof.
move => i.
rewrite /Di_def /Di_BPL.
case:(cast1 i, cast2 i).
rewrite !subr_addr => a b.
by rewrite castmx_id.
Qed.


Definition Di'_def (i:Z):=castmx(cast3 i,cast4 i)  (Di'_BPL(i-1-1)).

Definition Fi_def(i:Z):= castmx(cast2 i, cast4 i) (Fi_BPL (i-1-1)).

Definition Gi_def(i:Z):=castmx(cast4 i, cast2 i) (Gi_BPL(i-1-1)).
Definition Hi_def(i:Z):= castmx(cast2 i, cast1 i)(Hi_BPL(i-1-1)).



(*Proofs to build the chain complexes composed by both Di_def and Di'_def*)
Lemma ddi_def : forall i,  Di_def (i+1) *m Di_def i = 0.
Proof.
move=> i.
rewrite /Di_def.
have ->:castmx (cast1 (i + 1), cast2 (i + 1)) (Di_BPL (i + 1 - 1 - 1)) =
castmx (cast5 (i), cast1 (i)) (Di_BPL (i - 1 - 1 + 1)); last first.
  rewrite mulmx_cast; first  by rewrite !subr_addr.
  by rewrite ddi_BPL => H; rewrite  !castmx_const.
case: (cast1(i+1), cast2(i+1)).
case: (cast5 i, cast1 i).
rewrite !subr_addr !addr_subr /Di_BPL => a b a0 b0.
have -> : a=a0 by apply eq_irrelevance.
have ->: b=b0 by apply eq_irrelevance.
by done.
Qed.


Lemma ddi'_def : forall i,  Di'_def (i+1) *m Di'_def i = 0.
Proof.
move=> i.
rewrite /Di'_def /Di'_BPL.
have ->:castmx (cast3 (i + 1), cast4 (i + 1)) (Di'_BPL (i + 1 - 1 - 1))=
castmx (cast6 (i), cast3 (i)) (Di'_BPL (i - 1 - 1 + 1)); last first.
  rewrite mulmx_cast; first by rewrite !subr_addr.
  rewrite /Di'_BPL (nilpotency (D quasi_bpl) (i-1-1)).
  by move => H; rewrite  !castmx_const.
case: (cast3(i+1), cast4(i+1)).
case: (cast6 i, cast3 i).
rewrite !subr_addr !addr_subr /Di'_BPL=> a b a0 b0.
have -> : a=a0 by apply eq_irrelevance.
have ->: b=b0 by apply eq_irrelevance.
by done.
Qed.


(*Proofs to build the morphisms*)
Lemma df_fDi_def :forall i,  Di_def i *m Fi_def i =  Fi_def (i + 1) *m Di'_def i.
Proof.
move => i.
rewrite /Di_def /Fi_def /Di'_def.
have ->:  castmx (cast2 (i + 1), cast4 (i + 1)) (Fi_BPL (i + 1 - 1 - 1)) =
    castmx (cast1 i, cast3 (i)) (Fi_BPL (i - 1 - 1 + 1)); last first.
  rewrite !mulmx_cast; try by rewrite !subr_addr.
  rewrite df_fDi_BPL; move => [a b] [c d]. 
  have -> : a=c by apply eq_irrelevance.
  have ->: b=d by apply eq_irrelevance.
  by done.
case: (cast2(i+1), cast4(i+1)).
case: (cast1 i, cast3 i).
rewrite !subr_addr !addr_subr=> a b a0 b0.
have -> : a=a0 by apply eq_irrelevance.
have ->: b=b0 by apply eq_irrelevance.
by done.
Qed.


Lemma Dg_gdi_def : forall i,  Di'_def i *m Gi_def i = Gi_def (i+1) *m (Di_def i).
Proof.
move => i.
rewrite /Di_def /Gi_def /Di'_def.
have ->:  castmx ( cast4 (i + 1), cast2 (i + 1)) (Gi_BPL (i + 1 - 1 - 1)) =
    castmx ( cast3 (i), cast1 i) (Gi_BPL (i - 1 - 1 + 1)); last first.
 rewrite !mulmx_cast; try by rewrite !subr_addr.
  rewrite Dg_gdi_BPL; move => [a b] [c d]. 
  have -> : a=c by apply eq_irrelevance.
  have ->: b=d by apply eq_irrelevance.
  by done.
case: (cast4(i+1), cast2(i+1)).
case: (cast3 i, cast1 i).
rewrite !subr_addr !addr_subr=> a b a0 b0.
have -> : a=a0 by apply eq_irrelevance.
have ->: b=b0 by apply eq_irrelevance.
by done.
Qed.



(*Proofs to build a reduction*)
Lemma fgi_def : forall i, Gi_def i *m Fi_def i = 1%:M.
Proof.
move => i.
rewrite /Gi_def/Fi_def mulmx_cast; first by rewrite cast4.
rewrite (fgi_BPL(i-1-1)) => H.
by rewrite id_castmx.
Qed.


Lemma hgi_def : forall i, Gi_def i *m Hi_def i = 0.
Proof.
move => i.
rewrite /Gi_def/Hi_def mulmx_cast; first by rewrite cast1 cast4.
rewrite (hgi_BPL(i-1-1))=> h.
by rewrite castmx_const.
Qed.


Lemma hhi_def :forall i, Hi_def i *m Hi_def (i+1) = 0.
Proof.
move => i.
rewrite /Hi_def.
have ->:castmx (cast2 (i + 1), cast1 (i + 1)) (Hi_BPL (i + 1 - 1 - 1))=
castmx (cast1 (i), cast5 (i)) (Hi_BPL (i - 1 - 1 + 1)); last first.
  rewrite mulmx_cast; first by rewrite !subr_addr.
  by rewrite hhi_BPL; move => H; rewrite  !castmx_const.
case: (cast2(i+1), cast1(i+1)).
case: (cast1 i, cast5 i).
rewrite !subr_addr !addr_subr /Hi_BPL => a b a0 b0.
have -> : a=a0 by apply eq_irrelevance.
have ->: b=b0 by apply eq_irrelevance.
by done.
Qed.


Lemma fhi_def :forall i,  Hi_def i *m Fi_def (i+1) = 0.
Proof.
move => i.
rewrite /Fi_def/Hi_def.
have ->:  castmx (cast2 (i + 1), cast4 (i + 1)) (Fi_BPL (i + 1 - 1 - 1)) =
    castmx (cast1 i, cast3 (i)) (Fi_BPL (i - 1 - 1 + 1)); last first.
  rewrite mulmx_cast; first by rewrite !subr_addr.
  by rewrite fhi_BPL; move => H; rewrite  !castmx_const.
case: (cast2 (i+1), cast4 (i+1)).
case: (cast1 i, cast3 i).
rewrite !subr_addr !addr_subr /Hi_BPL => a b a0 b0.
have -> : a=a0 by apply eq_irrelevance.
have ->: b=b0 by apply eq_irrelevance.
by done.
Qed.



Lemma gf_Dh_hDi_def :forall i,  Fi_def (i+1) *m Gi_def (i+1) + Hi_def (i+1) *m Di_def (i+1)  + Di_def(i) *m Hi_def (i) = 1%:M.
Proof.
move => i.
rewrite /Di_def /Gi_def /Hi_def /Fi_def /Di'_def.
rewrite !mulmx_cast; try by rewrite !subr_addr.
move => [a b] [c d] [e f]. 
have ->: castmx (e, f) (Fi_BPL (i + 1 - 1 - 1) *m Gi_BPL (i + 1 - 1 - 1)) =
   castmx (a, b) (Fi_BPL (i - 1 - 1 +1) *m Gi_BPL (i - 1 - 1 +1)).
  case: (e, f).
  case: (a, b).
  rewrite !subr_addr !addr_subr => a0 b0 a1 b1.
  have -> : a0=a1 by apply eq_irrelevance.
  have ->: b0=b1 by  apply eq_irrelevance.
  by done.  
have ->: castmx (c, d) (Hi_BPL (i + 1 - 1 - 1) *m Di_BPL (i + 1 - 1 - 1)) =
   castmx (a, b) (Hi_BPL (i - 1 - 1 + 1) *m Di_BPL (i - 1 - 1 + 1)). 
  case: (c, d).
  case: (a, b).
  rewrite !subr_addr !addr_subr => a0 b0 a1 b1.
  have -> : a0=a1 by apply eq_irrelevance.
  have ->: b0=b1 by apply eq_irrelevance.
  by done. 
by rewrite !addmx_cast gf_Dh_hDi_BPL id_castmx.
Qed.

(*We build the structures and define the reduction*)

Definition Dii_def:= Build_FGChain_Complex ddi_def.
Definition Dii'_def:= Build_FGChain_Complex ddi'_def.

Definition f_def := Build_FGChain_Complex_Morphism (A:=Dii_def) (B:=Dii'_def) 
  df_fDi_def.

Definition g_def := Build_FGChain_Complex_Morphism (A:=Dii'_def) (B:=Dii_def) 
 Dg_gdi_def.

Definition h_def:= Build_FGHomotopy_operator (A:=Dii_def) Hi_def.

Definition rho_def := 
 (Build_FGReduction (F:=f_def)(G:=g_def)(H:=h_def)fgi_def gf_Dh_hDi_def fhi_def hgi_def hhi_def).



End definitions_def.


End perturbation.


End BPL.

