Add LoadPath "CoqEAL".
Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq choice fintype.
Require Import finfun tuple ssralg matrix mxalgebra bigop.
Require Import ZArith Zinfra.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.


(*The main structures are defined: 
  chaincomplex, morphism chain complex, homotopy operator, reduction*)


(*It's defined the notion of ker_min.
  The two main lemmas proved about ker_min are:
    
Lemma ker_min_kermx (m n : nat) (M : 'M_(m,n)) : 
  (kermx M :=: (ker_min M))%MS. 
  This means that kermx of a matrix generates the same space that the ker_min of
  this matrix. 
 
Lemma ker_intersection (m1 n1 n2: nat) (A:'M[K]_(m1,n1)) (B:'M[K]_(m1,n2)):
(kermx A :&: kermx B :=: kermx (row_mx A B))%MS.
  Moreover, the kermx of a row (A,B) is equal to the intersection of the kermx of A
  with the kermx of B.
*)


Section Z_Theory.


Lemma addr_subr (m n : Z) : (m + n - n)%Z = m.
Proof.
by apply/eqP; rewrite !GRing.subr_eq. 
Qed.

Lemma subr_addr (m n : Z) : (m - n + n)%Z = m.
Proof.
by rewrite -Z.sub_sub_distr -Zminus_diag_reverse -Zminus_0_l_reverse.
Qed. 

End Z_Theory.



Section Finitely_Generated_Chain_Complexes.

Local Open Scope ring_scope.
Variable K : fieldType.

Record FGChain_Complex :=
{
m : Z -> nat; (* number of generators in each dimension *)
diff : forall i:Z, 'M[K]_(m (i + 1), m i);
nilpotency : forall i:Z, (diff (i + 1)) *m (diff i) = 0 
}.

Record FGChain_Complex_Morphism (A B : FGChain_Complex) :=
{
M : forall i:Z, 'M[K]_((m A i),(m B i));
M_well_defined : forall i:Z, (diff A i) *m (M i) = (M (i+1)) *m (diff B i)
}.

Record FGHomotopy_operator (A : FGChain_Complex) :=
{Ho : forall i:Z, 'M[K]_(m A i, m A (i+1)%Z)}.

Section casts. 

Lemma cast_1_minus1 (C : FGChain_Complex) (i : Z): m C (i + 1 - 1) = m C i.
Proof.
by rewrite addr_subr.
Qed. 

Lemma cast_minus1_1 (C : FGChain_Complex) (i : Z): m C (i - 1 + 1) = m C i.
Proof.
by rewrite subr_addr.
Qed. 

End casts.


Record FGReduction := 
{ C : FGChain_Complex;
  D : FGChain_Complex;
  F : FGChain_Complex_Morphism C D;
  G : FGChain_Complex_Morphism D C;
  H : FGHomotopy_operator C;
  ax1 : forall i:Z, (M G i) *m (M F i) = 1%:M;
  ax2 : forall i:Z,  (M F (i+1)) *m (M G (i+1)) + 
   ((Ho H (i+1)) *m (diff C (i+1))) +
   ((diff C i) *m (Ho H i)
   )
   =  1%:M;
  ax3 : forall i:Z,  (Ho H i) *m (M F (i+1)) = 0;
  ax4 : forall i:Z, (M G i) *m (Ho H i) = 0;
  ax5 : forall i:Z, (Ho H i) *m (Ho H (i+1)) = 0
}.

End Finitely_Generated_Chain_Complexes.


Section definitions.

Open Local Scope ring_scope.

Variable K : fieldType.


Section auxiliary_results.

Lemma cast_rank1 (m n : nat) (M : 'M[K]_(m,n)) : 
 ((\rank M) + (m - (\rank M)) )%nat = m. 
Proof.
by rewrite (subnKC (rank_leq_row M)).
Qed.

Lemma ltnSadd (m n : nat) : m + n < m = false.
Proof.
elim : n => [|n IH]; first by rewrite addn0 ltnn.
rewrite addnS.
case H : ((m + n).+1 < m); last by done.
have H2 : m + n < (m + n).+1 by done.
move : (ltn_trans H2 H) => H3.
by rewrite H3 in IH.  
Qed.


Definition ker_min  (m n : nat) (M : 'M_(m,n)) :=
  (castmx ((Logic.eq_refl (m-\rank M)%nat), (cast_rank1 M))
          (row_mx (@const_mx _ (m-\rank M) (\rank M) 0) 
                  1%:M)) *m 
  (kermx M).



Lemma castmx_copid (m m' r: nat) (eq_mm : (m = m') * (m = m'))
  (R : ringType) :
  castmx eq_mm (copid_mx r) = (@copid_mx R m' r).
Proof. by case: eq_mm; case: m' / => b; rewrite castmx_id. Qed.

Lemma castmx_blockmx_copid (m m' r: nat) (eq_mm : ((r + m')%nat = m) * ((r+m')%nat = m))
  (R : ringType) : (castmx eq_mm (block_mx 0 0 0 1%:M)) = (@copid_mx R m r).
Proof.
case : eq_mm; case : m / => b; rewrite castmx_id.
apply/matrixP => i j; rewrite mxE; case : splitP => i0 H;
rewrite mxE; case : splitP => j0 Hj; 
rewrite !mxE {2}H ?(ltn_ord i0) ?Bool.andb_true_r ?ltnSadd 
        ?Bool.andb_false_r ?GRing.subr0 /=. 
      case  Hij : (@eq_op nat_eqType (@nat_of_ord (addn r m') i)
                    (@nat_of_ord (addn r m') j)).
         move : Hij; move/eqP => Hij;  move : (ord_inj Hij); 
         move/eqP => -> /=.
         have H11 : 1%:R == 1%:R by done.
         move : (H11 R) => H12.      
         rewrite -GRing.subr_eq0 in H12.
         by rewrite (eqP H12).
       case Hij2 : (i == j); last by rewrite GRing.subr0.
       move : (eqP Hij2) => Hij3.
       by rewrite Hij3 in Hij; move : Hij; move/eqP.
    case  Hij : (@eq_op nat_eqType (@nat_of_ord (addn r m') i)
                    (@nat_of_ord (addn r m') j)).
       move : Hij; move/eqP => Hij;  move : (ord_inj Hij); 
       move/eqP => -> /=.
       have H11 : 1%:R == 1%:R by done.
       move : (H11 R) => H12.      
       rewrite -GRing.subr_eq0 in H12.
       by rewrite (eqP H12).
     case Hij2 : (i == j); last by rewrite GRing.subr0.
     move : (eqP Hij2) => Hij3.
     by rewrite Hij3 in Hij; move : Hij; move/eqP.
  case Hij: (i == j); last by done.
  move : (eqP Hij) => Hij2.
  rewrite Hij2 in H.
  by move : (ltn_ord j0); rewrite -Hj H ltnSadd.
case Hij: (i0 == j0). 
  move : (eqP Hij) => Hij2.
  rewrite Hij2 in H.
  have Hij3 : (@eq_op nat_eqType (@nat_of_ord (addn r m') i)
                    (@nat_of_ord (addn r m') j)) by rewrite H Hj.
  move: (ord_inj (eqP Hij3)) => ->.
  by have -> : (j == j); by done. 
case Hij2 : (i == j); last by done.
move : (eqP Hij2) => Hij3.
rewrite Hij3 Hj in H.
have Hij4 : (r + j0 - r)%nat = (r + i0 - r)%nat by rewrite H. 
rewrite -{2 4}(addn0 r) !subn_add2l !subn0 in Hij4.
move : (ord_inj Hij4) => Hij5.
rewrite Hij5 in Hij.
by move : Hij; move/eqP.
Qed.


Lemma mulmx_cast : forall (R : ringType) (m1 n1 m2 n2 m3 m4 : nat) (eq_mn1 : (m1 = m2) * (n1 = n2))
     (eq_mn2 : (n1 = n2) * (m3 = m4)) (eq_mn3 : (m1 = m2) * (m3 = m4)) (A : 'M[R]_(m1, n1)) (B :'M[R]_(n1,m3)), 
     ((castmx eq_mn1 A) *m (castmx eq_mn2 B)) = (castmx eq_mn3 (A *m B)).
Proof.
move => R m1 n1 m2 n2 m3 m4 [eq_mn1 eq_mn1'] [eq_mn2 eq_mn2'] [eq_mn3 eq_mn3'] A B.
move: (eq_mn1) (eq_mn1') (eq_mn2) (eq_mn2') (eq_mn3) (eq_mn3') A B. 
rewrite eq_mn1 eq_mn1' eq_mn3' => eq_mn_0 eq_mn_1 eq_mn_2 eq_mn_3 eq_mn_4 eq_mn_5 A B.
by rewrite !castmx_id.
Qed.

Lemma ker_min_kermx (m n : nat) (M : 'M_(m,n)) : 
  (kermx M :=: (ker_min M))%MS. 
Proof.
apply/eqmxP/andP; split; apply/submxP; last first.
  by exists (castmx ((Logic.eq_refl (m-\rank M)%nat), (cast_rank1 M))
                    (row_mx (@const_mx _ (m-\rank M) (\rank M) 0) 
                            1%:M)).
exists (castmx ((cast_rank1 M),(Logic.eq_refl (m-\rank M)%nat))
               (col_mx (@const_mx _ (\rank M) (m-\rank M)%nat 0)
                       1%:M)).
rewrite /ker_min mulmxA mulmx_cast ?cast_rank1 //.
move => Hyp0.
by rewrite mul_col_row !mul0mx mulmx0 mulmx1 /kermx castmx_blockmx_copid 
           mulmxA copid_mx_id ?(rank_leq_row M). 
Qed.

Lemma ker_intersection (m1 n1 n2: nat) (A:'M[K]_(m1,n1)) (B:'M[K]_(m1,n2)):
(kermx A :&: kermx B :=: kermx (row_mx A B))%MS.
Proof.
apply/eqmxP; apply/rV_eqP => u.
rewrite sub_capmx !(sameP sub_kermxP eqP) mul_mx_row -row_mx0.
by  apply/andP/eqP=> [[/eqP-> /eqP->] | /eq_row_mx[-> ->]]. 
Qed.


End auxiliary_results.

End definitions.